@file:JvmName("ExtensionsKt")
@file:Suppress("SqlSourceToSinkFlow")
@file:OptIn(ExperimentalContracts::class)

package ru.landgrafhomyak.bgtu.db0.storage.sqlite_jdbc

import org.intellij.lang.annotations.Language
import java.sql.Connection
import java.sql.PreparedStatement
import java.sql.ResultSet
import kotlin.contracts.ExperimentalContracts
import kotlin.contracts.InvocationKind
import kotlin.contracts.contract


internal inline fun <R> Connection.transaction(t: (Connection) -> R): R {
	contract {
		callsInPlace(t, InvocationKind.EXACTLY_ONCE)
	}

	this.prepareStatement("BEGIN TRANSACTION").use { ps -> ps.executeUpdate() }
	val r: R
	var isSuccessful = true
	try {
		r = t(this)
	} catch (e1: Throwable) {
		isSuccessful = false
		try {
			this.prepareStatement("ROLLBACK TRANSACTION").use { ps -> ps.executeUpdate() }
		} catch (e2: Throwable) {
			e1.addSuppressed(e2)
		}
		throw e1
	} finally {
		if (isSuccessful)
			this.prepareStatement("COMMIT TRANSACTION").use { ps -> ps.executeUpdate() }
	}
	return r
}


internal fun Connection.executeStatement(@Language("SQL") sql: String) {
	this.prepareStatement(sql) { ps -> ps.execute() }
}


internal inline fun <R> Connection.prepareStatement(@Language("SQL") sql: String, action: (PreparedStatement) -> R): R {
	contract {
		callsInPlace(action, InvocationKind.EXACTLY_ONCE)
	}
	this.prepareStatement(sql).use { ps -> return action(ps) }
}

internal inline fun <R, C : MutableCollection<R>> ResultSet.mapTo(to: C, transform: (rs: ResultSet) -> R): C {
	while (this.next()) {
		to.add(transform(this))
	}
	return to
}

internal inline fun <R> ResultSet.map(transform: (rs: ResultSet) -> R): List<R> =
	this.mapTo(ArrayList(), transform)

internal inline fun <R, C : MutableCollection<R>> ResultSet.mapToThenClose(to: C, transform: (rs: ResultSet) -> R): C =
	this.use { rs -> rs.mapTo(to, transform) }

internal inline fun <R> ResultSet.mapThenClose(transform: (rs: ResultSet) -> R): List<R> =
	this.use { rs -> rs.map(transform) }

internal inline fun <R> PreparedStatement.executeQuery(action: (ResultSet) -> R): R {
	contract {
		callsInPlace(action, kotlin.contracts.InvocationKind.EXACTLY_ONCE)
	}
	return this.executeQuery().use(action)
}


internal fun PreparedStatement.setUByte(parameterIndex: Int, value: UByte): Unit =
	this.setByte(parameterIndex, value.toByte())

internal fun PreparedStatement.setUShort(parameterIndex: Int, value: UShort): Unit =
	this.setShort(parameterIndex, value.toShort())

internal fun PreparedStatement.setUInt(parameterIndex: Int, value: UInt): Unit =
	this.setInt(parameterIndex, value.toInt())

internal fun PreparedStatement.setULong(parameterIndex: Int, value: ULong): Unit =
	this.setLong(parameterIndex, value.toLong())

internal fun ResultSet.getUByte(columnIndex: Int): UByte =
	this.getByte(columnIndex).toUByte()

internal fun ResultSet.getUShort(columnIndex: Int): UShort =
	this.getShort(columnIndex).toUShort()

internal fun ResultSet.getUInt(columnIndex: Int): UInt =
	this.getInt(columnIndex).toUInt()

internal fun ResultSet.getULong(columnIndex: Int): ULong =
	this.getLong(columnIndex).toULong()