package ru.landgrafhomyak.bgtu.db0.telegram.impl

import kotlin.math.max
import org.json.JSONArray
import org.json.JSONObject
import ru.landgrafhomyak.bgtu.db0.telegram.api.CallbackQuery
import ru.landgrafhomyak.bgtu.db0.telegram.api.InlineButton
import ru.landgrafhomyak.bgtu.db0.telegram.api.Message
import ru.landgrafhomyak.bgtu.db0.telegram.api.TelegramApiException
import ru.landgrafhomyak.bgtu.db0.telegram.api.TelegramBotApi

class JavaBlockingHttpClientTelegramApiImpl(
	private val baseUrl: String,
	private val httpClient: JavaBlockingJson2JsonHttpClient
) : TelegramBotApi {
	constructor(
		token: String,
		baseUrlFactory: BaseUrlFactory = BaseUrlFactory.OfficialServer,
		httpClient: JavaBlockingJson2JsonHttpClient
	) : this(baseUrlFactory.bindWithBotToken(token), httpClient)


	fun interface BaseUrlFactory {
		fun bindWithBotToken(token: String): String

		companion object {
			@JvmStatic
			val tokenPattern = Regex("""\d+:[a-zA-Z0-9\-_]+""")
		}

		object OfficialServer : BaseUrlFactory {
			override fun bindWithBotToken(token: String): String {
				if (null == BaseUrlFactory.tokenPattern.matchEntire(token))
					throw IllegalArgumentException("Bad bot token")
				return "https://api.telegram.org/bot${token}/"
			}
		}
	}


	@Suppress("FunctionName")
	private inline fun <reified O> _doRequest(method: String, payload: JSONObject): O {
		val response = this.httpClient.request(this.baseUrl + method, payload)
		if (!response.getBoolean("ok")) {
			throw TelegramApiException(response.getString("description"))
		}

		@Suppress("LiftReturnOrAssignment")
		when (O::class) {
			Unit::class -> return Unit as O
			JSONObject::class -> return response.getJSONObject("result") as O
			JSONArray::class -> return response.getJSONArray("result") as O
			Boolean::class -> return response.getBoolean("result") as O
			String::class -> return response.getString("result") as O
			else -> throw IllegalArgumentException("Unsupported return type")
		}
	}

	@Suppress("UNCHECKED_CAST")
	override fun isChatOwner(chat: Long, user: Long): Boolean = this
		._doRequest<JSONArray>("getChatAdministrators", JSONObject(mapOf("chat_id" to chat)))
		.let { o -> o as Iterable<JSONObject> }
		.filter { jo -> jo.getString("status") == "creator" }
		.any { jo -> jo.getJSONObject("user").getLong("id") == user }


	@Suppress("FunctionName")
	private fun _parseUser(raw: JSONObject) = UserStruct(
		id = raw.getLong("id"),
		displayName = "${raw.optString("first_name")} ${raw.optString("last_name")}".trim()
	)

	@Suppress("FunctionName")
	private fun _parseMessage(raw: JSONObject): MessageStruct = MessageStruct(
		fromUser = this._parseUser(raw.getJSONObject("from")),
		chatId = raw.getJSONObject("chat").getLong("id"),
		messageId = raw.getLong("message_id"),
		_rawText = if (!raw.has("text")) null else raw.getString("text"),
		replyToMessage = if (!raw.has("reply_to_message")) null else this._parseMessage(raw.getJSONObject("reply_to_message")),
		inlineKeyboard = if (!raw.has("reply_markup")) null else raw.getJSONObject("reply_markup").let { kbd ->
			if (!kbd.has("inline_keyboard"))
				return@let null

			@Suppress("UNCHECKED_CAST")
			return@let kbd.getJSONArray("inline_keyboard")
				.let { rows -> rows as Iterable<Iterable<JSONObject>> }
				.map row@{ row ->
					return@row row.map btn@{ btn ->
						if (!btn.has("callback_data"))
							return@let null
						return@btn InlineButton(btn.getString("text"), btn.getString("callback_data"))
					}
				}
		}
	)

	@Suppress("FunctionName")
	private fun _formatInlineKeyboard(api: List<List<InlineButton>>): JSONObject = api
		.map { row -> row.map { btn -> JSONObject(mapOf("text" to btn.displayText, "callback_data" to btn.data)) }.let(::JSONArray) }
		.let { rr -> JSONObject(mapOf("inline_keyboard" to rr)) }

	override fun sendMessage(replyTo: Message, htmlText: String, buttons: List<List<InlineButton>>?): Message =
		this._parseMessage(
			this._doRequest<JSONObject>(
				"sendMessage", JSONObject(
					mapOf(
						"chat_id" to replyTo.chatId,
						"text" to htmlText,
						"parse_mode" to "html",
						"reply_markup" to buttons?.let(this::_formatInlineKeyboard)
					)
				)
			)
		)

	override fun editMessageText(message: Message, htmlText: String, keepButtons: Boolean): Message =
		this._parseMessage(
			this._doRequest<JSONObject>(
				"editMessageText", JSONObject(
					mapOf(
						"chat_id" to message.chatId,
						"message_id" to message.messageId,
						"text" to htmlText,
						"parse_mode" to "html",
						"reply_markup" to (
								if (!keepButtons) null
								else message.inlineKeyboard?.let(this::_formatInlineKeyboard)
								)
					)
				)
			)
		)

	override fun answerCallbackQuery(query: CallbackQuery, text: String) =
		this._doRequest<Unit>(
			"answerCallbackQuery", JSONObject(
				mapOf(
					"callback_query_id" to query.queryId,
					"text" to text
				)
			)
		)


	@Suppress("FunctionName")
	private fun _parseCallbackQuery(raw: JSONObject): CallbackQuery =
		CallbackQueryStruct(
			queryId = raw.getString("id"),
			user = this._parseUser(raw.getJSONObject("from")),
			message = this._parseMessage(raw.getJSONObject("message")),
			data = raw.getString("data")
		)

	fun interface UpdatesConsumer {
		fun processUpdate(u: IncomingUpdate)
	}

	@Suppress("UNCHECKED_CAST")
	fun receiveUpdatesUntilError(consumer: UpdatesConsumer) {
		var nextId = 0L
		while (true) {
			this
				._doRequest<JSONArray>(
					"getUpdates",
					JSONObject(mapOf("offset" to nextId, "timeout" to 1, "allowed_updates" to JSONArray(listOf("message", "callback_query"))))
				)
				.let { a -> a as Iterable<JSONObject> }
				.forEach { u ->
					nextId = max(nextId, u.getLong("update_id") + 1)
					if (u.has("message"))
						consumer.processUpdate(IncomingUpdate.TextMessage(this._parseMessage(u.getJSONObject("message"))))
					else if (u.has("callback_query"))
						consumer.processUpdate(IncomingUpdate.CallbackQuery(this._parseCallbackQuery(u.getJSONObject("callback_query"))))
				}
		}
	}
}